"""
This class is used to cache return value of functions on disk for a specified
number of days. This is used by lakshmi.assets module to cache name/ asset
value (i.e the slow functions). For examples on how to use this class, please
see the tests (tests/test_cache.py file).

Currently, this module can only be used on functions which are class members
and the function itself must take no arguments. These restrictions can be
easily relaxed, but so far that all usecases don't need anything more than what
is currently implemented.

In addition to caching values, this class also allows one to optionally call
a user-specified function on cache-misses (currently used to show a progress
bar to the user via the lak CLI).
"""

import concurrent.futures
import functools
import pickle
from abc import ABC, abstractmethod
from datetime import datetime
from hashlib import md5
from pathlib import Path

# Inspired by https://pypi.org/project/cache-to-disk/. I tried using other
# options such as requests-cache, but it was too slow compared to the solution
# implemented here.


class Cacheable(ABC):
    """Interface that declares that a particular class's method return
    values could be cached. The methods should not take a parameter,
    and cache_key() + method name should uniquely imply the return
    value of that class method."""
    @abstractmethod
    def cache_key(self):
        """Unique string value used as key for caching."""
        pass


def get_file_age(file):
    """Returns the age of file.

    Args:
        file: A PosixPath object representing a file.

    Returns: An int represeting the age in days.
    """
    return (datetime.today()
            - datetime.fromtimestamp(file.stat().st_mtime)).days


# Constants
# Default cache directory if none is specified.
_DEFAULT_DIR = Path.home() / '.lakshmicache'

_CACHE_STR = 'cache_dir'
_FORCE_STR = 'force_refresh'
_FORCED_FILES_STR = 'forced_files'
_MISS_FUNC_STR = 'miss_func'

# Dict (string -> object) to keep cache context.
# Description of keys to what is stored:
# _CACHE_STR:
# The pathlib.Path object specifying cache directory. If set to None,
# caching is disabled. Default: _DEFAULT_DIR
# _FORCE_STR:
# If set to True, new values are re-generated once even if a cached one is
# available. This is meant for data that is cached for < month (stock prices
# and Treasury Bond value). Values that are cached for > 40 days ignore this
# flag. Default: False
# _FORCED_FILES_STR:
# A set of files which are already refreshed once due to _ctx[_FORCE_STR]
# being set to True. this is used to ensure we don't re-fetch same values
# multiple times in a session.
# _MISS_FUNC_STR:
# If set, this function is called for every cache miss.
_ctx = {_FORCE_STR: False}


def set_force_refresh(v):
    """Sets whether cached values should be refreshed.

    Args:
        v: Boolean representing if cached values should be re-generated.
    """
    global _ctx
    _ctx[_FORCE_STR] = v
    _ctx[_FORCED_FILES_STR] = set()


def set_cache_miss_func(f):
    """Sets the function to call for cache-misses in case the cached function
    is directly called. This func is called periodically if prefetch() is
    used to fetch multiple cached values. This is useful for displaying
    progress bar while waiting for slow functions to complete.

    Args:
        f: The function to call whenever a cache-miss happens (i.e. whenever
        the underlying function is called instead of using a cached value) OR
        this function is called periodically when using prefetch to fetch
        multiple values in parallel.
    """
    global _ctx
    if f:
        _ctx[_MISS_FUNC_STR] = f
    else:
        # Clear out previously set function, if any.
        _ctx.pop(_MISS_FUNC_STR, None)


def set_cache_dir(cache_dir):
    """Sets the cache directory.

    If the cache directory is not specified, default ~/.lakshmicache
    is used.

    Args:
        cache_dir: The pathlib.Path object specifying cache directory.
        If set to None, caching is disabled.
    """
    global _ctx
    _ctx[_CACHE_STR] = cache_dir
    if cache_dir is None:
        return
    cache_dir.mkdir(exist_ok=True)  # Create cache dir if one doesn't exist.
    # Delete old files whose cache values are invalid already.
    for file in cache_dir.glob('*_*.lkc'):
        days = int(file.name.split('_')[0])
        if get_file_age(file) >= days:
            file.unlink()


def _valid_cached_value(file, days, add_to_ignored=True):
    """Helper function to check if the cached value from file is valid.

    Args:
        file: The Path object representing a file potentially containing
        previously cached value.
        days: Number of days after which the cached value becomes invalid.
        add_to_ignored: If force_refresh is set and this arg is set, the
        file is added to an ignored set of files so that this function doesn't
        return False for the same file if it is called again (this prevents
        refreshing the same cached value multiple times).

    Returns: True iff the cached value in file is valid.
    """
    MAX_DAYS_TO_FORCE_REFRESH = 40
    if (
        _ctx[_FORCE_STR]
        and days < MAX_DAYS_TO_FORCE_REFRESH
        and file.name not in _ctx[_FORCED_FILES_STR]
    ):
        # Ignore cached value.
        if add_to_ignored:
            _ctx[_FORCED_FILES_STR].add(file.name)
        return False
    return (file.exists() and get_file_age(file) < days)


def _call_func(class_obj, func):
    """Helper function to return value of class_obj.func().

    In addition to calling function, this helper also calls the
    cache_miss function if one is set in the context.

    Args:
        class_obj: The object of a particular class implementing Cacheable
        interface.
        func: The function whose return values has to be cached. Assumed
        to take no parameters.

    Returns: The return value of the func.
    """
    global _ctx
    if _MISS_FUNC_STR in _ctx:
        _ctx[_MISS_FUNC_STR]()
    return func(class_obj)


def _cache_filename(class_obj, func, days):
    """Returns the filename to be used for the args."""
    key = f'{func.__qualname__}_{class_obj.cache_key()}'
    return f'{days}_{md5(key.encode("utf8")).hexdigest()}.lkc'


def cache(days):
    """Returns decorator that caches functions return value on disk for
    specified number of days.

    Args:
        days: Number of days for which to cache the return value of the
        function.

    Returns: The decorator.
    """
    def decorator(func):
        @functools.wraps(func)
        def new_func(class_obj):
            global _ctx
            if _CACHE_STR not in _ctx:
                # Cache dir not set. Set to default.
                set_cache_dir(_DEFAULT_DIR)
            cache_dir = _ctx[_CACHE_STR]
            if not cache_dir:
                return _call_func(class_obj, func)

            file = cache_dir / _cache_filename(class_obj, func, days)
            if _valid_cached_value(file, days):
                return pickle.loads(file.read_bytes())
            value = _call_func(class_obj, func)
            file.write_bytes(pickle.dumps(value))
            return value
        new_func.cached_days = days
        return new_func
    return decorator


class _Prefetch:
    """Class to help with prefetching and caching of multiple values in
    parallel threads."""

    def __init__(self):
        self.cache_key_to_funcs = {}
        global _ctx
        if _CACHE_STR not in _ctx:
            # Cache dir not set. Set to default.
            set_cache_dir(_DEFAULT_DIR)
        self.cache_dir = _ctx[_CACHE_STR]

    def _return_cached_funcs(self, class_obj):
        all_methods = [getattr(class_obj, f) for f in dir(class_obj)]
        return [f for f in all_methods if callable(f)
                and hasattr(f, 'cached_days')]

    def add(self, class_obj):
        """Add class_obj to the list of objects whose cached methods are to be
        prefetched."""
        if not self.cache_dir:
            # Caching is disabled, don't prefetch.
            return

        cache_key = class_obj.cache_key()
        if self.cache_key_to_funcs.get(cache_key) is not None:
            # Already added to be prefetched.
            return

        funcs_to_refresh = []
        for func in self._return_cached_funcs(class_obj):
            file = self.cache_dir / _cache_filename(
                class_obj, func, func.cached_days)
            if not _valid_cached_value(file, func.cached_days,
                                       add_to_ignored=False):
                funcs_to_refresh.append(func)

        self.cache_key_to_funcs[cache_key] = funcs_to_refresh

    def fetch(self):
        """Fetch all cached methods of objects added earlier via add() call
        in parallel threads."""
        def prefetch_fn(funcs):
            for f in funcs:
                f()

        global _ctx
        # Reset cache miss func to None so it is not called from multiple
        # threads in parallel.
        cache_miss_func = None
        if _MISS_FUNC_STR in _ctx:
            cache_miss_func = _ctx[_MISS_FUNC_STR]
            set_cache_miss_func(None)

        fs = []
        with concurrent.futures.ThreadPoolExecutor() as executor:
            fs = [executor.submit(prefetch_fn, func_list) for func_list in
                  self.cache_key_to_funcs.values()]
            while len(fs) != 0:
                fs = concurrent.futures.wait(fs, timeout=0.1).not_done
                if cache_miss_func:
                    cache_miss_func()

        # Reset the map, so it can be optionally used again with add() method.
        self.cache_key_to_funcs = {}
        # Restore cache miss funcion.
        set_cache_miss_func(cache_miss_func)


# Global object of type Prefetch used for prefetching cached functions using
# parallel threads.
_prefetch_obj = None


def prefetch_add(class_obj):
    """Add class_obj to list of objects who cached methods are to be
    prefetched."""
    global _prefetch_obj
    if _prefetch_obj is None:
        _prefetch_obj = _Prefetch()
    _prefetch_obj.add(class_obj)


def prefetch():
    """Fetch all cached methods of objects added earlier via the prefetch_add()
    call in parallel threads."""
    assert _prefetch_obj is not None, (
        'prefetch_add must be called before calling prefetch')
    _prefetch_obj.fetch()



"""Classes to analyze the portfolio.

These classes are not meant to change the portfolio. They are used for
just analyzing the portfolio and output the results (if any).
"""

from abc import ABC, abstractmethod

import numpy as np

from lakshmi.table import Table
from lakshmi.utils import format_money


# TODO(sarvjeets): This is not needed/used anywhere. Consider removing it.
class Analyzer(ABC):
    @abstractmethod
    def analyze(self, portfolio):
        pass


class TLH(Analyzer):
    """Tax loss harvesting opportunity analyzer.

    This triggers a message with details of tax lots that could be sold to
    harvest the losses.
    """

    def __init__(self, max_percentage, max_dollars=None):
        """Triggers if a lot has lost more than max_percentage or the total loss
        for a given asset exceeds max_dollars.

        Args:
            max_percentage: A float representing the max percentage loss after
            which a lot can be harvested.
            max_dollars: A float representing the max amount of loss (in
            dollars) for one or more lots (of the same security), beyond which
            we should recommend selling all the lots that are below their
            cost basis.
        """
        assert max_percentage > 0.0 and max_percentage < 1.0, (
            'max_percetage should be between 0% and 100% (exclusive).')
        self.max_percentage = max_percentage
        if max_dollars:
            assert max_dollars > 0, 'max_dollars should be positive.'
        self.max_dollars = max_dollars

    def _return_lots_to_sell(self, price, tax_lots):
        """Helper function that returns which lots to harvest.

        Args:
            price: The current price of the security.
            tax_lots: A list of lakshmi.assets.TaxLot (all belonging
            to the same security).

        Returns: A list of [lot date, loss in dollar, loss in percent]
        that could be loss harvested.
        """
        percent_lots = []
        negative_lots = []
        total_loss = 0.0

        for lot in tax_lots:
            loss = (lot.unit_cost - price) * lot.quantity
            loss_percent = (lot.unit_cost - price) / lot.unit_cost
            if loss > 0:
                negative_lots.append([lot.date, loss, loss_percent])
                total_loss += loss
            if loss_percent > self.max_percentage:
                percent_lots.append([lot.date, loss, loss_percent])

        if self.max_dollars is not None and total_loss > self.max_dollars:
            return negative_lots
        else:
            return percent_lots

    def analyze(self, portfolio):
        """Returns which lots can be tax loss harvested from a portfolio.

        Args:
            portfolio: A lakshmi.Portfolio object representing a portfolio.

        Returns: A lakshmi.table.Table object with tax lots to harvest with
        columns corespnding to 'Account', 'Asset', 'Date', 'Loss' and 'Loss%'.
        """
        ret_val = Table(
            5,
            headers=['Account', 'Asset', 'Date', 'Loss', 'Loss%'],
            coltypes=['str', 'str', 'str', 'dollars', 'percentage'])
        for account in portfolio.accounts():
            for asset in account.assets():
                if hasattr(asset, 'get_lots') and asset.get_lots():
                    for lot in self._return_lots_to_sell(
                            asset.price(), asset.get_lots()):
                        ret_val.add_row(
                            [account.name(), asset.short_name()] + lot)
        return ret_val


class BandRebalance(Analyzer):
    """Triggers if portfolio asset class targets are outside the bands
    specified. This considers an asset class outside bound if the absolute
    difference in percentage allocation is more than max_abs_percent different
    from target allocation or more than max_relative_percent different from the
    target allocation.

    A popular version is the 5/25 band based rebalancing rule. For more
    information, please see 5/25 on:
    https://www.bogleheads.org/wiki/Rebalancing
    """

    def __init__(self, max_abs_percent=0.05, max_relative_percent=0.25):
        """Constructor to set the bands. An asset class is considered outside
        the rebalance bands if the differnt between the desired and actual
        ratio of that asset class exceeds the lessor of max absolute or max
        relative percentage.

        Args:
            max_abs_percent: A float (ratio) representing the maximum absolute
            percent that an asset can deviate before it's considered outside
            the rebalancing band.
            max_relative_percent: A float (ratio) representing the maximum
            relative percent that an asset can deviate before it's considered
            outside the rebalancing band.
        """
        assert max_abs_percent > 0 and max_abs_percent < 1.0
        assert max_relative_percent > 0 and max_relative_percent < 1.0
        self.max_abs_percent = max_abs_percent
        self.max_relative_percent = max_relative_percent

    def analyze(self, portfolio):
        """Returns asset classes that are outside the rebalancing bands.

        Args:
            portfolio: A lakshmi.Portfolio object representing the portfolio
            that is being analyzed.

        Returns: A lakshmi.table.Table object with columns representing
        Asset Class, Actual % of asset class in portfolio, Desired % of asset
        class in the portfolio, Value in dollars of the asset class and the
        difference between Desired and actual dollar values.
        """
        aa = portfolio.asset_allocation(portfolio.asset_classes.leaves())
        headers = ['Class', 'Actual%', 'Desired%', 'Value', 'Difference']
        assert headers == aa.headers()
        ret_val = Table(
            5,
            headers,
            ['str', 'percentage', 'percentage', 'dollars', 'delta_dollars'])
        for row in aa.list():
            abs_percent = abs(row[1] - row[2])
            rel_percent = abs_percent / row[2] if row[2] != 0 else 0
            if (
                abs_percent >= self.max_abs_percent
                or rel_percent >= self.max_relative_percent
            ):
                ret_val.add_row(row)

        return ret_val


class _Solver:
    """Internal class to help allocate cash to portfolio (helper class for the
    next class).

    I tried using scipy optimize instead of this custom solver and it was
    extremely flaky for this purpose (even after scaling/conditioning the
    inputs well). So finally, I came up theis heuristic algo which I 'feel'
    works, but I haven't proved it formally.  This class can be further
    optimized/fixed, but for now it serves its purpose.  -- sarvjeets

    Notation:
    - f_i: Money in asset i orginally.
    - x_i: New money to be added to asset i (what we are solving for).
    - A_j: Money in asset class j.
    - C_j: Money in asset class j before adding new money (implementd as
    self.money)
    - d_j: Desired ratio of asset class j (implemented as self.desired_ratio).
    - a_{ij}: Ratio of asset i in asset class j (implemented as
    self.allocation).

    So,
    A_j = \\sum_i a_{ij} (f_i + x_i)
    \\sum_j A_j = T (a constant, given by self.total)

    Error E = \\sum_j (A_j/(d_j T) - 1)^2

    Derivative wrt a asset i (implemented as self.derivative)
    dE/dx_i = (2/T) \\sum_j (a_{ij} / d_j) (A_j/(d_j T) - 1)
    Rate of change of above derivative wrt asset k:
    d^2E/{dx_i dx_k} = (2/T^2) \\sum_j (a_{ij} a_{kj} / d^2_j)

    The algorithm (very loose description here, if this works, I'll probably
    write it more formally):
    0. Without lost of generality, assume cash to be allocated is positive (
    the other way round just replaces min with max in the following algo). We
    also dedup same assets so that the equations in Step 2 has a unique
    solution.
    1. Start by the lowest derivative asset.
    2. Add money to asset to match the derivative of a target asset. Pick the
    next target asset based on whichever target asset leads to lowest
    derivative.
    3. Now we have two assets with same derivatives, keep repeating step 2
    until all assets have same derivatives.
    4. If we run out of cash to allocate before all assets have same
    derivative, exit early.
    5. If we still have money after all assets have same derivative, add extra
    money to all assets while keeping their derivatives the same.

    There is special case where we run out of money in an asset. In that case,
    we just zero out the asset, take it from the list of assets that are
    being optimized and go back to step 2.

    For step 2, we solve the following linear equations:
    k is summing over assets with same derivatives. n is target asset to which
    we are not adding money (aka x_n = 0).

    For each i:
    dE/dx_i + \\sum_{k != i} x_k d^2E/{dx_i dx_k} =
                 dE/dx_n + \\sum_k x_k d^2E/{dx_i dx_k}
    (This equation is solved in self.compute_delta).

    self.allocate_all_cash solves:
    dE/dx_i + \\sum_{k != i} x_k d^2E/{dx_i dx_k} =
                 dE/dx_i (without any additional money allocated to it) + 0.1
    which simplifies to:
    for each i:
    \\sum_k (x_k \\sum_j a{ij} a_{kj} / d^2_j = 0.1T
    """

    def __init__(self, aa, assets, cash, total):
        """
        Args:
            aa: Map of asset class name to a tuple of desired_ratio and
            money allocated to it.
            assets: List of lakshmi.Assets to which to allocate the money to.
            cash: The cash to be allocated.
            total: The total value of the portfolio (T in equation above).
        """
        self.aa = aa
        self.total = total
        self.cash = cash
        self.assets = assets
        # Used to pick either the min or max gradient.
        self.best_gradient_fn = np.argmin if cash > 0 else np.argmax
        self.adjusted_values = [asset.adjusted_value() for asset in assets]

    def desired_ratio(self, asset_class):
        """Helper function to give desired ratio of an asset class."""
        return self.aa[asset_class][0]

    def money(self, asset_class):
        """Helper function to give money allocated to an asset class."""
        return self.aa[asset_class][1]

    def asset_classes(self):
        """Returns list of asset classes."""
        return self.aa.keys()

    def allocation(self, i, j):
        """Returns allocation of asset i in asset class j."""
        return self.assets[i].class2ratio.get(j, 0.0)

    def update_aa(self, new_money):
        """Updates self.aa with new_money.

        Args:
            new_money: A list of deltas to be which new_money[i] is for
            self.assets[i].
        """
        for asset, money in zip(self.assets, new_money):
            for ac, ratio in asset.class2ratio.items():
                self.aa[ac] = self.aa[ac][0], self.aa[ac][1] + ratio * money

    def derivative(self, i):
        """Computes dE/dx_i. The actual derivative has (2/T) extra
        term, but as it is common in all assets, we ignore it (the rest of the
        computations in equations are done correctly to account for a missing
        2/T factor).
        """
        sum = 0.0
        for j in self.asset_classes():
            rel_ratio = (self.money(j) / (self.desired_ratio(j) * self.total)
                         - 1)
            sum += rel_ratio * (self.allocation(i, j) / self.desired_ratio(j))
        return sum

    def compute_delta(self, source_assets, target_asset):
        """Heart of this solver. This computes deltas (x) on the source_
        assets, so that the derivative of the error function wrt source_assets
        is equal to the target_asset.

        Args:
            source_assets: A set of indices referring to self.assets
            representing assets on which to compute deltas on.
            target_asset: An index referring to an asset in self.assets. The
            delta is computed for source_assets to make their error
            derivative equal to the target_asset.

        Returns: A tuple of computed deltas and the final derivative. The
        derivative of target asset can change based on the deltas computed.
        """
        a = []
        b = []
        n = target_asset
        alpha = self.derivative(n)

        for i in source_assets:
            equation_i = []
            for k in source_assets:
                coeff = 0.0
                for j in self.asset_classes():
                    coeff += (
                        (self.allocation(k, j) / self.desired_ratio(j) ** 2)
                        * (self.allocation(i, j) - self.allocation(n, j)))
                equation_i.append(coeff)
            a.append(equation_i)
            const_i = self.total * alpha
            for j in self.asset_classes():
                const_i += (
                    (self.allocation(i, j) / self.desired_ratio(j) ** 2)
                    * (self.desired_ratio(j) * self.total - self.money(j)))
            b.append(const_i)

        solution = np.linalg.lstsq(a, b, rcond=None)[0].tolist()
        x = np.zeros(len(self.assets))

        final_derivative = alpha
        for k in source_assets:
            x[k] = solution.pop(0)
            sum = 0.0
            for j in self.asset_classes():
                sum += (self.allocation(n, j) * self.allocation(k, j)
                        / self.desired_ratio(j) ** 2)
            final_derivative += sum * x[k] / self.total

        return x, final_derivative

    def allocate_all_cash(self, cash_to_allocate, equal_gradient_assets):
        """Allocates all the remaining cash. This function assumes that
        all the derivatives of the error function wrt assets are the same.
        Then it solves for extra deltas in each asset that will increase
        the derivative by 0.1 (arbitrary number). It then re-scales this
        number to make sure that the new deltas equal to cash_to_allocate.
        The exact computation is listed in the class level comment.

        Args:
            cash_to_allocate: The cash to allocate.

        Returns: deltas (x), one for each self.asset that sums up to
        cash_to_allocate and ensures the final derivatives of error function
        wrt all assets are equal.
        """
        a = []
        for i in equal_gradient_assets:
            equation_i = []
            for k in equal_gradient_assets:
                coeff = 0.0
                for j in self.asset_classes():
                    coeff += (self.allocation(i, j) * self.allocation(k, j)
                              / self.desired_ratio(j) ** 2)
                equation_i.append(coeff)
            a.append(equation_i)
        x = np.linalg.lstsq(a, [0.1 * self.total] * len(equal_gradient_assets),
                            rcond=None)[0]
        x *= cash_to_allocate / np.sum(x)

        ret_val = [0] * len(self.assets)
        for i, j in zip(equal_gradient_assets,
                        range(len(equal_gradient_assets))):
            ret_val[i] = x[j]

        return ret_val

    def bound_at_zero(self, x, deltas, equal_gradient_assets, zeroed_assets):
        """Ensure that none of the solution exceeds the available money in
        assets. If that happens, it zeros out the asset, removes the asset
        from the set of assets that are being optmized (equal_gradient_assets).
        It additinally adds the funds to zeroed_assets.

        Args:
            x: The current solution.
            deltas: The new delta on the solution that is being considered to
            be added to x.
            equal_gradient_assets: Assets that have equal error gradients.
            zeroed_assets: Assets that have zero balance.

        Returns:
            None, if applying (x+deltas) to assets would not cause their
            balance to become negative.
            new_deltas, a list of len(self.assets), otherwise. new_deltas
            is computed to zero out money in any asset which would have
            gotton a negative balance if (x+deltas) was applied to it.
        """
        if self.cash > 0:
            # in this case, we don't have to worry about bounding the money,
            # as we only add money to assets and never remove it.
            return None

        new_deltas = [0] * len(deltas)
        adjusted = False

        for i in equal_gradient_assets:
            money_removed = -(x[i] + deltas[i])
            if money_removed > self.adjusted_values[i]:
                # Withdrew too much money.
                new_deltas[i] = -(self.adjusted_values[i] + x[i])
                zeroed_assets.add(i)
                adjusted = True
            else:
                new_deltas[i] = 0

        equal_gradient_assets -= zeroed_assets
        return new_deltas if adjusted else None

    def solve(self):
        """The main method. This method implements the algorithm listed in the
        class comment.
        """
        # Pick the lowest error gradient asset.
        equal_gradient_assets = set([self.best_gradient_fn(
            [self.derivative(i) for i in range(len(self.assets))])])
        zeroed_assets = set({})
        x = np.zeros(len(self.assets))
        left_cash = self.cash

        # Keep equalizing and adding an asset to equal_gradient_assets.
        while ((len(equal_gradient_assets)
                + len(zeroed_assets) != len(self.assets))
               and abs(left_cash) > 1e-6):
            # Handle the case when we zero out all equal_gradient_assets.
            if len(equal_gradient_assets) == 0:
                equal_gradient_assets.add(self.best_gradient_fn(
                    [self.derivative(i) for i in range(len(self.assets)) if
                     i not in zeroed_assets]))

            derivatives = []
            deltas = []
            indices = []
            # Compute the min increase in gradient among the remaining assets.
            for n in set(range(len(self.assets))) - equal_gradient_assets:
                delta, derivative = self.compute_delta(
                    equal_gradient_assets, n)
                derivatives.append(derivative)
                deltas.append(delta)
                indices.append(n)

            # Pick the highest or lowest derivative.
            best_asset_index = self.best_gradient_fn(derivatives)
            best_delta = deltas[best_asset_index]
            new_cash = np.sum(best_delta)

            if abs(new_cash) >= abs(left_cash):
                # We have allocated more cash than what was left.
                best_delta = best_delta * left_cash / new_cash
                new_cash = left_cash

            new_delta = self.bound_at_zero(
                x, best_delta, equal_gradient_assets, zeroed_assets)
            if new_delta:
                best_delta = new_delta
                new_cash = np.sum(new_delta)
            x += best_delta
            self.update_aa(best_delta)
            left_cash -= new_cash
            if not new_delta:  # No zeroed assets.
                equal_gradient_assets.add(indices[best_asset_index])

        # Handle any left over cash.
        while abs(left_cash) > 1e-6:
            delta = self.allocate_all_cash(left_cash, equal_gradient_assets)
            new_delta = self.bound_at_zero(x, delta, equal_gradient_assets,
                                           zeroed_assets)
            if new_delta:
                delta = new_delta
            x += delta
            self.update_aa(delta)
            left_cash -= np.sum(delta)

        return x


class Allocate(Analyzer):
    """Allocates any unallocated cash in the account to assets.

    If an account has any unallocated cash (aka what if) then this class
    allocates that cash to the assets in the account. The allocation is done
    with the goal of minimizing the relative ratio of actual allocation of
    asset classes to the desired allocation. Cash could be negative in which
    case money is removed from the assets.

    The allocation to assets is done via lakshmi.Portfolio.what_if function;
    and hence can be reset easily (by calling
    lakshmi.Portfolio.reset.what_ifs).
    """

    def __init__(self, account_name, exclude_assets=[], rebalance=False):
        """
        Args:
            account_name: The full name of the account to analyze.
            exclude_assets: A list of asset short names (strings) which
            will not be allocated any new cash from the account as a result
            of calling analyze.
            rebalance: If False, money is either only added (in case cash is
            positive) or removed (in case cash is negative) from the assets.
            If set, money is added and removed (as needed) from assets
            to minimize the relative difference from the desired asset
            allocation.
        """
        self.account_name = account_name
        self.exclude_assets = exclude_assets
        self.rebalance = rebalance

    def _apply_whatifs(self, portfolio, assets, deltas, saved_whatifs=None):
        """Apply whatifs given by deltas to assets in the portfolio."""
        table = Table(2, ['Asset', 'Delta'], ['str', 'delta_dollars'])
        if not saved_whatifs:
            saved_whatifs = [0] * len(assets)
        for asset, delta, saved in zip(assets, deltas, saved_whatifs):
            portfolio.what_if(self.account_name,
                              asset.short_name(),
                              float(delta))  # delta could be a np.float
            table.add_row([asset.short_name(), delta - saved])
        return table

    def analyze(self, portfolio):
        """Modifies portfolio by allocating any cash and returns the resulting
        changes.

        Args:
            portfolio: The portfolio on which to operate on. This portfolio is
            modified by applying what_ifs to the assets in the provided
            account.

        Returns:
            A table.Table of asset names and an delta for each of the asset.

        Throws:
            AssertionError: In case of
            - There is no cash to allocation and rebalance is False.
            - An asset class's desired allocation ratio is zero.
            - No assets are present in the Account after taking out the
            exlcuded assets.
            - Cash to withdraw is more than the total value of assets in the
            portfolio.
            - For some reason, we can't minimize the difference between
            relative error of actual vs desired allocation, given all the
            constraints.
        """
        account = portfolio.get_account(self.account_name)
        cash = account.available_cash()
        assert cash != 0 or self.rebalance, (
            f'No available cash to allocate in {self.account_name}.')
        assets = [x for x in account.assets() if x.short_name() not in
                  self.exclude_assets]
        assert len(assets) != 0, 'No assets to allocate cash to.'

        saved_whatifs = None
        if self.rebalance:
            saved_whatifs = []
            # Withdraw all cash and reallocate.
            for asset in assets:
                saved_whatifs += [asset.adjusted_value()]
                portfolio.what_if(self.account_name, asset.short_name(),
                                  -asset.adjusted_value())
            cash = account.available_cash()

        total = sum([asset.adjusted_value() for asset in assets])
        if abs(total + cash) < 1e-6:
            # Withdraw all cash and return.
            return self._apply_whatifs(
                portfolio, assets,
                [-asset.adjusted_value() for asset in assets])
        assert -cash < total, (
            f'Cash to withdraw ({format_money(-cash)}) is more than the total '
            f'available money in the account ({format_money(total)}).')

        # Map of leave asset classes -> (desired%, money)
        asset_allocation = {}
        for row in portfolio.asset_allocation(
                portfolio.asset_classes.leaves()).list():
            asset_allocation[row[0]] = row[2], row[3]

        for asset in assets:
            for ac in asset.class2ratio.keys():
                assert asset_allocation[ac][0] != 0, (
                    f'Desired ratio of asset class {ac} cannot be zero.')

        result = _Solver(
            asset_allocation, assets, cash, portfolio.total_value()).solve()
        return self._apply_whatifs(portfolio, assets, result, saved_whatifs)



"""This module contains all classes and functions related to checkpointing
and computing portfolio's performance."""

import bisect
from dataclasses import dataclass
from datetime import datetime, timedelta

import yaml
from pyxirr import xirr

from lakshmi import utils
from lakshmi.table import Table


class Checkpoint:
    """Class representing a single checkpoint of the portfolio. Each checkpoint
    represents a single day. The checkpoint contains the portfolio value, and
    money inflows and outflows on that day.
    """

    def __init__(self, checkpoint_date, portfolio_value, inflow=0, outflow=0):
        """Constructs a new checkpoint for the given date.

        Args:
            checkpoint_date: Date in 'YYYY/MM/DD' format.
            portfolio_value: The value of the portfolio on that date.
            inflow: The amount of money flowing into the portfolio on date.
            outflow: The amount of money flowing out of the portfolio on date.
        """
        self._date = utils.validate_date(checkpoint_date)

        assert portfolio_value > 0, 'Portfolio value must be positive'
        assert inflow >= 0, 'Inflow must be non-negative'
        assert outflow >= 0, 'Outflow must be non-negative'

        # Round portfolio value to 2 degits.
        self._portfolio_value = round(portfolio_value, 2)
        self._inflow = inflow
        self._outflow = outflow

    def to_dict(self, show_empty_cashflow=False, show_date=True):
        """Converts this object to a dictionary.

        Args:
            show_empty_cashflow: If set to True, inflows and outflows are
            shown even if they are empty.
            show_date: If set to False, date is not included in the dictionary
            output.

        Returns:
            A dictionary object representing this object.
        """
        d = {}
        if show_date:
            d['Date'] = self._date
        d['Portfolio Value'] = self._portfolio_value
        if show_empty_cashflow or self._inflow > 0:
            d['Inflow'] = self._inflow
        if show_empty_cashflow or self._outflow > 0:
            d['Outflow'] = self._outflow
        return d

    @classmethod
    def from_dict(cls, d, date=None):
        """Returns a new object given dictionary representation.

        This function is reverse of the above function. Optionally, it takes
        in a date which is used to set the date if it is specified. If it is
        not specified than this function expects that a data is specified via
        d. This is a reverse function of to_dict().

        Args:
            d: A dictionary representing this object (usually a output of
            to_dict method).
            date: Date for this checkpoint (Format: YYYY/MM/DD).

        Returns: A newly initialized object of type Checkpoint.

        Raises:
            AssertionError if the dictionary format is not as expected.
        """
        if not date:
            date = d.pop('Date')
        ret_obj = Checkpoint(date, d.pop('Portfolio Value'),
                             d.pop('Inflow', 0), d.pop('Outflow', 0))
        assert len(d) == 0, f'Extra attributes found while parsing: {list(d)}'
        return ret_obj

    def get_date(self):
        """Returns date of this checkpoint in 'YYYY/MM/DD' format."""
        return self._date

    def get_portfolio_value(self):
        """Returns the checkpoint's portfolio value."""
        return self._portfolio_value

    def get_inflow(self):
        """Returns the money flowing in."""
        return self._inflow

    def get_outflow(self):
        """Returns the money flowing out."""
        return self._outflow


class Timeline:
    """Class representing a collection of checkpoints."""

    _DATE_FMT = '%Y/%m/%d'

    def __init__(self, checkpoints):
        """Returns a new object given a list of checkpoints."""
        assert len(checkpoints) > 0

        self._checkpoints = {}
        self._dates = []
        for cp in checkpoints:
            cp_date = cp.get_date()
            assert cp_date not in self._dates, (
                f'Cannot have two checkpoints with the same date ({cp_date})')
            self._dates.append(cp_date)
            self._checkpoints[cp_date] = cp
        self._dates.sort()

    def to_list(self):
        """Returns this object as a list of checkpoints."""
        return [self._checkpoints[date].to_dict() for date in self._dates]

    @classmethod
    def from_list(cls, timeline_list):
        """Returns a new object given a list (reverse of method above)."""
        return Timeline([Checkpoint.from_dict(cp) for cp in timeline_list])

    def to_table(self, begin=None, end=None):
        """Convert this timeline to a Table.

        This function is useful for pretty-printing this object.

        Args:
            begin: If specified, start printing checkpoints from this date
            (inclusive). Format: 'YYYY/MM/DD'. If None, start from the earliest
            checkpoint date.
            end: If specified, stop printing checkpoints after this date
            (inclusive). Format: 'YYYY/MM/DD'. If None, end at the last
            checkpoint date.

        Returns: A lakshmi.table.Table object.
        """
        table = Table(4,
                      headers=['Date', 'Portfolio Value', 'Inflow', 'Outflow'],
                      coltypes=['str', 'dollars', 'dollars', 'dollars'])
        begin_pos = bisect.bisect_left(self._dates, begin) if begin else None
        end_pos = bisect.bisect_right(self._dates, end) if end else None
        for date in self._dates[begin_pos:end_pos]:
            cp = self._checkpoints[date]
            table.add_row([cp.get_date(),
                           cp.get_portfolio_value(),
                           cp.get_inflow(),
                           cp.get_outflow()])
        return table

    def has_checkpoint(self, date):
        """Retuns true iff there is a checkpoint for date."""
        return utils.validate_date(date) in self._checkpoints

    def begin(self):
        """Returns the beginnning date of this timeline."""
        return self._dates[0]

    def end(self):
        """Returns the end date of this timeline."""
        return self._dates[-1]

    def covers(self, date):
        """Returns true if date is within the timeline."""
        date = utils.validate_date(date)
        return (date >= self.begin() and date <= self.end())

    @classmethod
    def _interpolate_checkpoint(cls, date, checkpoint1, checkpoint2):
        """Given checkpoints 1 and 2, returns new checkpoint for date."""
        date1 = datetime.strptime(checkpoint1.get_date(), Timeline._DATE_FMT)
        date2 = datetime.strptime(checkpoint2.get_date(), Timeline._DATE_FMT)
        given_date = datetime.strptime(date, Timeline._DATE_FMT)
        val1 = checkpoint1.get_portfolio_value()
        val2 = (checkpoint2.get_portfolio_value()
                - checkpoint2.get_inflow()
                + checkpoint2.get_outflow())

        interpolated_value = val1 + (val2 - val1) * (
            (given_date - date1) / (date2 - date1))
        return Checkpoint(date, interpolated_value)

    def get_checkpoint(self, date, interpolate=False):
        """Returns checkpoint for a given date.

        This function will return the checkpoint for date if it already exists
        in the Timeline. If there is no checkpoint for date, this function
        throws an error if interpolate is False. If interpolate is set to True
        this function will linearly interpolate the portfolio values around
        the given date and return a newly created checkpoint with that
        calculated value.

        Args:
            date: The date (in 'YYYY/MM/DD' format) for which to return the
            checkpoint.
            interpolate: If True, computes and returns a checkpoint for a date
            even if one doesn't exists in the Timeline.
        Returns:
            Checkpoint object corresponding to date.
        Raises:
            AssertionError, if interpolate is False and there is no checkpoint
            for date.
            AssertError, If date is not without the range of checkpoints in
            this timeline.
        """
        date = utils.validate_date(date)

        if self.has_checkpoint(date):
            return self._checkpoints[date]

        # date is not one of the saved checkpoints...

        assert interpolate, f'{date} is not one of the saved checkpoints.'

        # ... and it's OK to interpolate

        assert self.covers(date), (
            f'{date} is not in the range of the saved checkpoints. '
            f'Begin={self.begin()}, End={self.end()}')

        pos = bisect.bisect(self._dates, date)
        return Timeline._interpolate_checkpoint(
            date,
            self._checkpoints[self._dates[pos - 1]],
            self._checkpoints[self._dates[pos]])

    def insert_checkpoint(self, checkpoint, replace=False):
        """Inserts given checkpoint to the timeline.

        Args:
            checkpoint: Checkpoint to insert.
            replace: If True, the new checkpoint overwrites a checkpoint with
            the same date present in the timeline.

        Raises:
            AssertionError if replace is False and another checkpoint with the
            same date is present in the timeline.
        """
        date = checkpoint.get_date()

        if replace and self.has_checkpoint(date):
            self._checkpoints[date] = checkpoint
            return

        assert not self.has_checkpoint(date), (
            f'Cannot insert two checkpoints with the same date ({date}).')
        pos = bisect.bisect(self._dates, date)
        self._dates.insert(pos, date)
        self._checkpoints[date] = checkpoint

    def delete_checkpoint(self, date):
        """Removes checkpoint corresponding to date."""
        date = utils.validate_date(date)
        assert date in self._checkpoints
        self._checkpoints.pop(date)
        self._dates.remove(date)

    @dataclass
    class PerformanceData:
        # List of dates (datetime objects). Used to compute XIRR.
        dates: list
        # List of cashflows on the above dates. Money flowing out of portfolio
        # is considered positive. Used to compute XIRR.
        amounts: list
        # Beginning balance.
        begin_balance: float
        # Ending balance.
        end_balance: float
        # Sum of all inflows to the portfolio.
        inflows: float = 0.0
        # Sum of all outflows from the portfolio.
        outflows: float = 0.0

    def get_performance_data(self, begin, end):
        """Returns data in a format to help calculate XIRR.

        Args:
            begin: Begin date in YYYY/MM/DD format. If None, start from the
            earliest checkpoint date.
            end: End date in YYYY/MM/DD format. If None, ends at the last
            checkpoint date.

        Returns: PerformanceData object.
        """
        if not begin:
            begin = self.begin()
        if not end:
            end = self.end()

        assert utils.validate_date(begin) != utils.validate_date(end)

        dates = []
        amounts = []
        inflows = 0.0
        outflows = 0.0

        begin_checkpoint = self.get_checkpoint(begin, True)
        dates.append(datetime.strptime(begin_checkpoint.get_date(),
                                       Timeline._DATE_FMT))
        amounts.append(-begin_checkpoint.get_portfolio_value())

        begin_pos = bisect.bisect_right(self._dates, begin)
        end_pos = bisect.bisect_left(self._dates, end)
        for date in self._dates[begin_pos:end_pos]:
            dates.append(datetime.strptime(date, Timeline._DATE_FMT))
            checkpoint = self._checkpoints[date]
            amounts.append(checkpoint.get_outflow() - checkpoint.get_inflow())
            inflows += checkpoint.get_inflow()
            outflows += checkpoint.get_outflow()

        end_checkpoint = self.get_checkpoint(end, True)
        dates.append(
            datetime.strptime(end_checkpoint.get_date(), Timeline._DATE_FMT))
        amounts.append(end_checkpoint.get_portfolio_value()
                       + end_checkpoint.get_outflow()
                       - end_checkpoint.get_inflow())
        inflows += end_checkpoint.get_inflow()
        outflows += end_checkpoint.get_outflow()
        return Timeline.PerformanceData(
            dates=dates, amounts=amounts, inflows=inflows, outflows=outflows,
            begin_balance=begin_checkpoint.get_portfolio_value(),
            end_balance=end_checkpoint.get_portfolio_value())


class Performance:
    """Class to compute performance stats given a Timeline object."""

    _TIME_PERIODS = [timedelta(days=30),
                     timedelta(days=30) * 3,
                     timedelta(days=30) * 6,
                     timedelta(days=365),
                     timedelta(days=365) * 3,
                     timedelta(days=365) * 10]
    _TIME_PERIODS_NAMES = ['1 Month',
                           '3 Months',
                           '6 Months',
                           '1 Year',
                           '3 Years',
                           '10 Years']

    def __init__(self, timeline):
        self._timeline = timeline

    def get_timeline(self):
        """Returns the timeline."""
        return self._timeline

    def to_dict(self):
        """Converts this object to a dictionary."""
        return {'Timeline': self._timeline.to_list()}

    @classmethod
    def from_dict(cls, d):
        """Creates a new Performance object from a dict (reverse of above)."""
        ret_obj = Performance(Timeline.from_list(d.pop('Timeline')))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def save(self, filename):
        """Save this Object to a file."""
        with open(filename, 'w') as f:
            yaml.dump(self.to_dict(), f)

    @classmethod
    def load(cls, filename):
        """Load Performance object from a file."""
        with open(filename) as f:
            return Performance.from_dict(
                yaml.load(f.read(), Loader=yaml.SafeLoader))

    def _get_periods(self):
        """Returns periods for which summary stats should be printed."""
        # We only show 3 _TIME_PERIODS based on timeline_period
        timeline_period = (
            datetime.strptime(self._timeline.end(), Timeline._DATE_FMT)
            - datetime.strptime(self._timeline.begin(), Timeline._DATE_FMT))
        end_index = bisect.bisect_left(Performance._TIME_PERIODS,
                                       timeline_period)
        begin_index = max(0, end_index - 3)
        return (Performance._TIME_PERIODS[begin_index:end_index],
                Performance._TIME_PERIODS_NAMES[begin_index:end_index])

    @classmethod
    def _create_summary_row(cls, period_name, perf_data):
        """Helper method to create a row in summary table."""
        change = perf_data.end_balance - perf_data.begin_balance
        return [period_name, perf_data.inflows, perf_data.outflows,
                change, change / perf_data.begin_balance,
                xirr(perf_data.dates, perf_data.amounts)]

    def summary_table(self):
        """Returns summary of performance during different periods."""
        table = Table(6,
                      headers=['Period', 'Inflows', 'Outflows',
                               'Portfolio Change', 'Change %', 'IRR'],
                      coltypes=['str', 'dollars', 'dollars', 'delta_dollars',
                                'percentage_1', 'percentage_1'])
        # Not enough data for any points.
        if self._timeline.begin() == self._timeline.end():
            return table

        # Add rows for atmost 3 periods.
        periods, period_names = self._get_periods()
        for period, period_name in zip(periods, period_names):
            begin_date_str = (
                datetime.strptime(self._timeline.end(), Timeline._DATE_FMT)
                - period).strftime(Timeline._DATE_FMT)
            table.add_row(Performance._create_summary_row(
                period_name, self._timeline.get_performance_data(
                    begin_date_str, self._timeline.end())))

        # Add row for 'Overall' time period
        table.add_row(Performance._create_summary_row(
            'Overall', self._timeline.get_performance_data(
                self._timeline.begin(), self._timeline.end())))
        return table

    def get_info(self, begin, end):
        """Get information about the performance in [begin, end].

        This method prints detailed information about performance of portfolio
        (as given by checkpoints).

        Args:
            begin: Begin date in YYYY/MM/DD format. If None, starts at the
            earliest checkpoint date.
            end: End date in YYYY/MM/DD format. If None, ends at the last
            checkpoint date.

        Returns: A formatted string suitable for pretty-printing.
        """
        # Make dates strings cannonical.
        begin = utils.validate_date(begin) if begin else self._timeline.begin()
        end = utils.validate_date(end) if end else self._timeline.end()
        if begin == end:
            # Not enought checkpoints to compute performance.
            return ''

        data = self._timeline.get_performance_data(begin, end)
        change = data.end_balance - data.begin_balance

        table = Table(2, coltypes=['str', 'str'])
        growth = round(100 * change / data.begin_balance, 1)
        irr = round(100 * xirr(data.dates, data.amounts), 1)
        table.set_rows([
            ['Start date', begin],
            ['End date', end],
            ['Beginning balance', utils.format_money(data.begin_balance)],
            ['Ending balance', utils.format_money(data.end_balance)],
            ['Inflows', utils.format_money(data.inflows)],
            ['Outflows', utils.format_money(data.outflows)],
            ['Portfolio growth', utils.format_money_delta(change)],
            ['Market growth', utils.format_money_delta(
                change - data.inflows + data.outflows)],
            ['Portfolio growth %', f'{growth}%'],
            ['Internal Rate of Return', f'{irr}%']])
        return table.string(tablefmt='plain')



"""Common utils for Lakshmi."""

from datetime import datetime


def format_money(x):
    """Formats input (money) to a string.

    For example, if x=5.238, the output is '$5.24'.

    Args:
        x: Float (non-negative) representing dollars.
    """
    return '${:,.2f}'.format(x)


def format_money_delta(x):
    """Formats input (money delta) into a string.

    For example, if x=-23.249m the output is '-$23.25'.

    Args:
        x: Float (postive or negative) representating dollars.
    """
    return '{}${:,.2f}'.format('-' if x < 0 else '+', abs(x))


def validate_date(date_text):
    """Validates if the date is in the YYYY/MM/DD format.

    This function either throws a ValueError or returns the date_text
    formatted according to YYYY/MM/DD format.

    Args:
        date_text: Date text to be validated.

    Returns:
        Correctly formatted string representing date_text date.

    Throws:
        ValueError if date is not in YYYY/MM/DD format.
    """
    return datetime.strptime(date_text, '%Y/%m/%d').strftime('%Y/%m/%d')
